import Redis from 'ioredis';
import { LockHandle } from './queue-worker';
export interface RedisQueueWireMessage {
    t: number;
    c: number;
    s: number;
    d: any;
}
export interface HandleMessageCallArguments {
    data: any;
    context: {
        lock: LockHandle;
        timestamp: number | null;
        producer: number | null;
        sequence: number | null;
        latencyMs: number | null;
    };
}
export declare type HandleMessageCall = ({ data, context: { lock } }: HandleMessageCallArguments) => Promise<void>;
export interface ConstructorArgs {
    redis: Redis;
    batchSize: number;
    groupVisibilityTimeoutMs: number;
    pollingTimeoutMs: number;
    consumerCount: number;
    redisKeyPrefix: string;
}
export declare class RedisQueueClient {
    private redis;
    private batchSize;
    private groupVisibilityTimeoutMs;
    private pollingTimeoutMs;
    private consumerCount;
    private groupStreamKey;
    private groupSetKey;
    private consumerGroupId;
    private messagePriorityQueueKeyPrefix;
    private consumers;
    private clientId;
    private lastMessageSequenceNumber;
    private isRunning;
    private _ensured_stream_group;
    private callGetMetrics;
    private callAddGroupAndMessageToQueue;
    private statTotalInvalidMessagesCount;
    private statLastMessageLatencies;
    private statLastMessageReceiveTimestamps;
    private _handleMessageStats;
    constructor({ redis, batchSize, groupVisibilityTimeoutMs, pollingTimeoutMs, consumerCount, redisKeyPrefix }: ConstructorArgs);
    private _ensureStreamGroup;
    getMetrics({ topMessageGroupsLimit }: {
        topMessageGroupsLimit?: number | undefined;
    }): Promise<any>;
    startConsumers(handleMessage: HandleMessageCall, handleInvalidMessage?: HandleMessageCall): Promise<void>;
    stopConsumers(): Promise<void>;
    send({ data, priority, groupId }: {
        data: any;
        priority: number;
        groupId: string;
    }): Promise<number>;
    dispose(): Promise<void>;
    private _createPriorityMessageQueueKey;
}
