"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisQueueClient = void 0;
const redis_utils_1 = require("./redis-utils");
const queue_worker_1 = require("./queue-worker");
const scripts = __importStar(require("./lua-scripts"));
;
;
;
class RedisQueueClient {
    constructor({ redis, batchSize = 10, groupVisibilityTimeoutMs = 60000, pollingTimeoutMs = 10000, consumerCount = 10, redisKeyPrefix = '{redis-ordered-queue}' }) {
        this.consumers = [];
        this.clientId = 0;
        this.lastMessageSequenceNumber = 0;
        this._ensured_stream_group = false;
        this.callGetMetrics = null;
        this.callAddGroupAndMessageToQueue = null;
        this.statTotalInvalidMessagesCount = 0;
        this.statLastMessageLatencies = [];
        this.statLastMessageReceiveTimestamps = [];
        this.redis = redis;
        this.batchSize = batchSize;
        this.groupVisibilityTimeoutMs = groupVisibilityTimeoutMs;
        this.pollingTimeoutMs = pollingTimeoutMs;
        this.consumerCount = consumerCount;
        this.groupStreamKey = `${redisKeyPrefix}::msg-group-stream`;
        this.groupSetKey = `${redisKeyPrefix}::msg-group-set`;
        this.consumerGroupId = `${redisKeyPrefix}::consumer-group`;
        this.messagePriorityQueueKeyPrefix = `${redisKeyPrefix}::msg-group-queue::`;
        this.consumers = [];
        this.isRunning = false;
        this._ensured_stream_group = false;
    }
    _handleMessageStats(wireMessage) {
        const now = Date.now();
        const latencyMs = Math.max(0, now - wireMessage.t);
        this.statLastMessageLatencies.push(latencyMs);
        while (this.statLastMessageLatencies.length > 100) {
            this.statLastMessageLatencies.unshift();
        }
        this.statLastMessageReceiveTimestamps.push(now);
        while (this.statLastMessageReceiveTimestamps.length > 100) {
            this.statLastMessageReceiveTimestamps.unshift();
        }
    }
    _ensureStreamGroup() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._ensured_stream_group)
                return;
            this.clientId = (yield this.redis.call('CLIENT', ['ID']));
            // Ensure consumer groupId exists on the processing groupId stream
            try {
                yield this.redis.call('XGROUP', ['CREATE', this.groupStreamKey, this.consumerGroupId, 0, 'MKSTREAM']);
                this._ensured_stream_group = true;
            }
            catch (_a) {
            }
        });
    }
    getMetrics({ topMessageGroupsLimit = 10 }) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.callGetMetrics) {
                this.callGetMetrics = yield (0, redis_utils_1.prepare_redis_script)(this.redis, scripts.GetMetrics);
            }
            // Ensure consumer groupId exists on the processing groupId stream
            yield this._ensureStreamGroup();
            const response = yield this.callGetMetrics([this.consumerGroupId, topMessageGroupsLimit], [this.groupStreamKey, this.groupSetKey]);
            function transformList(result, keyName, valueName) {
                const out = [];
                for (let i = 0; i < result.length; i += 2) {
                    out.push({
                        [keyName]: result[i],
                        [valueName]: parseInt(result[i + 1], 10)
                    });
                }
                return out;
            }
            let sumLatency = 0;
            for (const latency of this.statLastMessageLatencies) {
                sumLatency += latency;
            }
            const avgLatency = this.statLastMessageLatencies.length > 0 ? sumLatency / this.statLastMessageLatencies.length : 0;
            return {
                bufferedMessageGroups: response[0],
                trackedMessageGroups: response[1],
                workingConsumers: response[2],
                visibleMessages: response[3],
                invalidMessages: this.statTotalInvalidMessagesCount,
                minLatencyMs: this.statLastMessageLatencies.length ? Math.min(...this.statLastMessageLatencies) : 0,
                maxLatencyMs: this.statLastMessageLatencies.length ? Math.max(...this.statLastMessageLatencies) : 0,
                avgLatencyMs: Math.round(avgLatency),
                topMessageGroups: transformList(response[4], 'group', 'backlog'),
            };
        });
    }
    startConsumers(handleMessage, handleInvalidMessage) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isRunning)
                throw new Error('Queue consumer is already running');
            this.isRunning = true;
            const callClaimTimedOutGroup = yield (0, redis_utils_1.prepare_redis_script)(this.redis, scripts.ClaimTimedOutGroup);
            const callUnlockGroup = yield (0, redis_utils_1.prepare_redis_script)(this.redis, scripts.UnlockGroup);
            const callDeleteMessage = yield (0, redis_utils_1.prepare_redis_script)(this.redis, scripts.DeleteMessage);
            // Ensure consumer groupId exists on the processing groupId stream
            yield this._ensureStreamGroup();
            for (let consumerIndex = 0; consumerIndex < this.consumerCount; ++consumerIndex) {
                const consumer = new queue_worker_1.RedisQueueWorker({
                    redis: this.redis,
                    batchSize: this.batchSize,
                    groupVisibilityTimeoutMs: this.groupVisibilityTimeoutMs,
                    pollingTimeoutMs: this.pollingTimeoutMs,
                    // Message handling callback
                    handleMessage: (msgData, lock) => __awaiter(this, void 0, void 0, function* () {
                        let wireMessage = null;
                        try {
                            wireMessage = JSON.parse(msgData.toString());
                        }
                        catch (_a) { }
                        if (wireMessage && !!wireMessage.d && !!wireMessage.t && !!wireMessage.c && !!wireMessage.s) {
                            this._handleMessageStats(wireMessage);
                            return yield handleMessage({
                                data: wireMessage.d,
                                context: {
                                    timestamp: wireMessage.t,
                                    producer: wireMessage.c,
                                    sequence: wireMessage.s,
                                    latencyMs: Math.max(0, Date.now() - wireMessage.t),
                                    lock
                                }
                            });
                        }
                        else {
                            ++this.statTotalInvalidMessagesCount;
                            // Bad message format
                            if (handleInvalidMessage) {
                                return yield handleInvalidMessage({
                                    data: msgData,
                                    context: { lock, timestamp: null, producer: null, sequence: null, latencyMs: null }
                                });
                            }
                            else {
                                console.error('Unhandled invalid message received: ', msgData);
                            }
                        }
                    }),
                    // Redis keys
                    groupStreamKey: this.groupStreamKey,
                    groupSetKey: this.groupSetKey,
                    consumerGroupId: this.consumerGroupId,
                    // Redis script EVALSHA hashes
                    callClaimTimedOutGroup,
                    callUnlockGroup,
                    callDeleteMessage,
                    // Priority queues key prefix generator
                    createPriorityMessageQueueKey: this._createPriorityMessageQueueKey.bind(this)
                });
                this.consumers.push(consumer);
            }
            this.consumers.map(consumer => consumer.start());
        });
    }
    stopConsumers() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isRunning) {
                yield Promise.all(this.consumers.map(consumer => consumer.stop()));
                this.consumers = [];
                this.isRunning = false;
            }
        });
    }
    send({ data, priority = 1, groupId }) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.callAddGroupAndMessageToQueue) {
                this.callAddGroupAndMessageToQueue = yield (0, redis_utils_1.prepare_redis_script)(this.redis, scripts.AddGroupAndMessageToQueue);
            }
            // Ensure consumer groupId exists on the processing groupId stream
            yield this._ensureStreamGroup();
            const wireMessage = {
                t: Date.now(),
                c: this.clientId,
                s: ++this.lastMessageSequenceNumber,
                d: data
            };
            const [addedMsgCount] = yield this.callAddGroupAndMessageToQueue([groupId, priority, JSON.stringify(wireMessage)], [this.groupStreamKey, this.groupSetKey, this._createPriorityMessageQueueKey(groupId)]);
            return addedMsgCount;
        });
    }
    dispose() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.stopConsumers();
        });
    }
    _createPriorityMessageQueueKey(groupId) {
        return this.messagePriorityQueueKeyPrefix + groupId;
    }
}
exports.RedisQueueClient = RedisQueueClient;
;
//# sourceMappingURL=queue-client.js.map